#!/user/bin/env python
'''structureToProteinDimers.py:

Maps a structure to its protein dimers

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from mmtf.utils import *
from mmtf.api.mmtf_writer import MMTFEncoder
import itertools
from sympy import Point3D
from mmtfPyspark.utils import DistanceBox
import time
import numpy as np
import math


class StructureToProteinDimers(object):
    '''Maps a protein structure to it's protein dimers

    Attributes
    ----------
        cutoffDistance (float): cutoff distance for protein dimers [8.0]
        contacts (int): number of contacts [20]
        useAllAtoms (bool): flag to use all atoms [False]
        exclusive (bool): exclusive flag [False]
    '''

    def __init__(self, cutoffDistance=8.0, contacts=20,
                 useAllAtoms=False, exclusive=False):
        self.cutoffDistance = cutoffDistance
        self.contacts = contacts
        self.useAllAtoms = useAllAtoms
        self.exclusive = exclusive

    def __call__(self, t):
        structure = t[1]

        # split structure into a list of chains
        chains = self._split_to_chains(structure)
        chainVectors = self._get_chain_vectors(chains)
        resList = []

        if self.useAllAtoms:
            boxes = self._get_all_atoms_distance_boxes(
                chains, self.cutoffDistance)
        else:
            boxes = self._get_c_beta_atoms_distance_boxes(
                chains, self.cutoffDistance)

        self.exclusiveHashSet = np.empty([0, 3])

        for i in range(len(chains)):

            j = 0

            while j < i:

                if self._check_pair(boxes[i], boxes[j], chains[i], chains[j], self.cutoffDistance, self.contacts):

                    if self.exclusive:

                        newVec = chainVectors[i] - chainVectors[j]
                        if not self._check_list(newVec, self.exclusiveHashSet):

                            resList.append(self._combine_chains(
                                chains[i], chains[j]))

                            self.exclusiveHashSet = np.append(
                                self.exclusiveHashSet, [newVec], axis=0)
                    else:
                        resList.append(self._combine_chains(
                            chains[i], chains[j]))

                j += 1

        return resList

    def _check_list(self, vec, exclusiveList):

        for point in exclusiveList:

            if np.linalg.norm(vec - point) < 0.1 and self._angle(vec, point) < 0.1:
                return True
            vec = vec * -1

            if np.linalg.norm(vec - point) < 0.1 and self._angle(vec, point) < 0.1:
                return True
            vec = vec * -1

        return False

    def _angle(self, a, b):

        arccosInput = np.dot(a, b) / np.linagl.norm(a) / np.linagl.norm(b)
        arccosInput = 1.0 if arccosInput > 1.0 else arccosInput
        arccosInput = -1.0 if arccosInput < -1.0 else arccosInput

        return math.acos(arccosInput)

    def _get_chain_vectors(self, chains):

        chainVectors = []

        for chain in chains:

            chainVectors.append(self._calc_average_vec(chain))

        return chainVectors

    def _calc_average_vec(self, s1):

        totX, totY, totZ = 0, 0, 0

        for i in range(s1.num_atoms):  # TODO double check num atoms

            totX += s1.x_coord_list[i]
            totY += s1.y_coord_list[i]
            totZ += s1.z_coord_list[i]

        return np.array([totX / s1.num_atoms, totY / s1.num_atoms, totZ / s1.num_atoms])

    def _distance(self, s1, s2, index1, index2):

        xCoord = s1.x_coord_list[index1]
        yCoord = s1.y_coord_list[index1]
        zCoord = s1.z_coord_list[index1]

        newPoint1 = Point3D(xCoord, yCoord, zCoord)

        xCoord = s2.x_coord_list[index2]
        yCoord = s2.y_coord_list[index2]
        zCoord = s2.z_coord_list[index2]

        newPoint2 = Point3D(xCoord, yCoord, zCoord)

        return newPoint1.distance(newPoint2)

    def _check_pair(self, box1, box2, s1, s2, cutoffDistance, contacts):

        pointsInBox2 = box1.getIntersection(box2)
        pointsInBox1 = box2.getIntersection(box1)

        hs1, hs2 = set(), set()

        num = 0

        for i in range(len(pointsInBox2)):

            for j in range(len(pointsInBox1)):

                if (i in hs1) or (j in hs2):
                    continue

                if self._distance(s1, s2, pointsInBox2[i], pointsInBox1[j]) < cutoffDistance:

                    num += 1
                    hs1.add(i)
                    hs2.add(j)

                if num > contacts:
                    return True

        return False

    def _get_c_beta_atoms_distance_boxes(self, chains, cutoffDistance):
        '''Get distance boxes for all atoms
        '''
        distanceBoxes = []

        for i in range(len(chains)):

            tmp = chains[i]
            newbox = DistanceBox(cutoffDistance)
            groupIndex = 0
            atomIndex = 0

            for k in range(tmp.groups_per_chain[0]):
                groupType = tmp.group_type_list[groupIndex]

                for m in range(len(tmp.group_list[groupType]["formalChargeList"])):

                    atomName = tmp.group_list[groupType]["atomNameList"][m]

                    if atomName == "CB":

                        xCoord = tmp.x_coord_list[atomIndex]
                        yCoord = tmp.y_coord_list[atomIndex]
                        zCoord = tmp.z_coord_list[atomIndex]

                        # TODO trying numpy instead of Point3D
                        #newPoint = Point3D(xCoord, yCoord, zCoord)
                        newPoint = np.array([xCoord, yCoord, zCoord])

                        newbox.add_point(newPoint, atomIndex)

                    atomIndex += 1

                groupIndex += 1

            distanceBoxes.append(newbox)

        return distanceBoxes

    def _get_all_atoms_distance_boxes(self, chains, cutoffDistance):
        '''Get distance boxes for all atoms
        '''
        distanceBoxes = []

        for i in range(len(chains)):

            tmp = chains[i]
            newbox = DistanceBox(cutoffDistance)

            for j in range(tmp.num_atoms):

                xCoord = tmp.x_coord_list[j]
                yCoord = tmp.y_coord_list[j]
                zCoord = tmp.z_coord_list[j]

                # TODO trying numpy instead of Point3D
                # newPoint = Point3D(xCoord, yCoord, zCoord)
                newPoint = np.array([xCoord, yCoord, zCoord])

                newbox.add_point(newPoint, j)

            distanceBoxes.append(newbox)

        return distanceBoxes

    def _split_to_chains(self, s):
        '''split structure to a list of chains
        '''
        chains = []
        numChains = s.chains_per_model[0]

        chainToEntityIndex = self._get_chain_to_entity_index(s)
        atomsPerChain, bondsPerChain = self._get_num_atoms_and_bonds(s)

        groupCounter = 0
        atomCounter = 0

        for i in range(numChains):
            atomMap = {}
            newChain = MMTFEncoder()
            entityToChainIndex = chainToEntityIndex[i]

            structureId = s.structure_id + '.' +\
                s.chain_name_list[i] + '.' +\
                s.chain_id_list[i] + '.' +\
                str(entityToChainIndex + 1)

            # Set header
            newChain.init_structure(bondsPerChain[i],
                                    atomsPerChain[i],
                                    s.groups_per_chain[i],
                                    1, 1, structureId)
            decoder_utils.add_xtalographic_info(s, newChain)
            decoder_utils.add_header_info(s, newChain)

            # Set model info (only one model: 0)
            newChain.set_model_info(0, 1)

            # Set entity and chain info
            newChain.set_entity_info([0],
                                     s.entity_list[entityToChainIndex]['sequence'],
                                     s.entity_list[entityToChainIndex]['description'],
                                     s.entity_list[entityToChainIndex]['type'])

            newChain.set_chain_info(s.chain_id_list[i],
                                    s.chain_name_list[i],
                                    s.groups_per_chain[i])

            for j in range(s.groups_per_chain[i]):
                groupIndex = s.group_type_list[groupCounter]
                # print(s.group_type_list)

                # Set group info
                newChain.set_group_info(s.group_list[groupIndex]['groupName'],
                                        s.group_id_list[groupCounter],
                                        s.ins_code_list[groupCounter],
                                        s.group_list[groupIndex]['chemCompType'],
                                        len(s.group_list[groupIndex]
                                            ['atomNameList']),
                                        len(s.group_list[groupIndex]
                                            ['bondOrderList']),
                                        s.group_list[groupIndex]['singleLetterCode'],
                                        s.sequence_index_list[groupCounter],
                                        s.sec_struct_list[groupCounter])

                for k in range(len(s.group_list[groupIndex]['atomNameList'])):
                    newChain.set_atom_info(s.group_list[groupIndex]['atomNameList'][k],
                                           s.atom_id_list[atomCounter],
                                           s.alt_loc_list[atomCounter],
                                           s.x_coord_list[atomCounter],
                                           s.y_coord_list[atomCounter],
                                           s.z_coord_list[atomCounter],
                                           s.occupancy_list[atomCounter],
                                           s.b_factor_list[atomCounter],
                                           s.group_list[groupIndex]['elementList'][k],
                                           s.group_list[groupIndex]['formalChargeList'][k])

                    atomCounter += 1

                for l in range(len(s.group_list[groupIndex]['bondOrderList'])):

                    bondIndOne = s.group_list[groupIndex]['bondAtomList'][l * 2]
                    bondIndTwo = s.group_list[groupIndex]['bondAtomList'][l * 2 + 1]
                    bondOrder = s.group_list[groupIndex]['bondOrderList'][l]

                    #newChain.set_group_bond(bondIndOne, bondIndTwo, bondOrder)
                    newChain.current_group.bond_atom_list += [
                        bondIndOne, bondIndTwo]
                    newChain.current_group.bond_order_list.append(bondOrder)

                groupCounter += 1

                # TODO skipping adding inter group bond info for now

            newChain.finalize_structure()

            # TODO double check if just getting from chain 0
            chain_type = [chain['type'] for chain in newChain.entity_list]

            polymer = "polymer" in chain_type

            if polymer:

                match = True

                for j in range(newChain.groups_per_chain[0]):

                    #print(newChain.groups_per_chain, newChain.group_type_list)
                    groupIndex = newChain.group_type_list[j]

                    if match:
                        _type = newChain.group_list[groupIndex]["chemCompType"]

                        match = (
                            _type == "L-PEPTIDE LINKING") or (_type == "PEPTIDE LINKING")

                if match:
                    chains.append(newChain)

        return chains

    def _combine_chains(self, s1, s2):

        if not s1.alt_loc_set:
            s1 = s1.set_alt_loc_list()
        if not s2.alt_loc_set:
            s2 = s2.set_alt_loc_list()

        groupCounter = 0
        atomCounter = 0

        structureId = s1.structure_id + "_append_" + s2.structure_id

        combinedStructure = MMTFEncoder()

        # Set header
        combinedStructure.init_structure(s1.num_bonds + s2.num_bonds,
                                         s1.num_atoms + s2.num_atoms,
                                         s1.num_groups + s2.num_groups,
                                         2, 1, structureId)
        decoder_utils.add_xtalographic_info(s1, combinedStructure)
        decoder_utils.add_header_info(s1, combinedStructure)

        # Set model info (only one model: 0)
        combinedStructure.set_model_info(0, 2)

        chainToEntityIndex = self._get_chain_to_entity_index(s1)[0]

        # Set entity and chain info
        combinedStructure.set_entity_info([0],
                                          s1.entity_list[chainToEntityIndex]['sequence'],
                                          s1.entity_list[chainToEntityIndex]['description'],
                                          s1.entity_list[chainToEntityIndex]['type'])

        combinedStructure.set_chain_info(s1.chain_id_list[0],
                                         s1.chain_name_list[0],
                                         s1.groups_per_chain[0])

        for i in range(s1.groups_per_chain[0]):
            groupIndex = s1.group_type_list[groupCounter]

            # Set group info
            combinedStructure.set_group_info(s1.group_list[groupIndex]['groupName'],
                                             s1.group_id_list[groupCounter],
                                             s1.ins_code_list[groupCounter],
                                             s1.group_list[groupIndex]['chemCompType'],
                                             len(s1.group_list[groupIndex]
                                                 ['atomNameList']),
                                             len(s1.group_list[groupIndex]
                                                 ['bondOrderList']),
                                             s1.group_list[groupIndex]['singleLetterCode'],
                                             s1.sequence_index_list[groupCounter],
                                             s1.sec_struct_list[groupCounter])

            for j in range(len(s1.group_list[groupIndex]['atomNameList'])):
                combinedStructure.set_atom_info(s1.group_list[groupIndex]['atomNameList'][j],
                                                s1.atom_id_list[atomCounter],
                                                s1.alt_loc_list[atomCounter],
                                                s1.x_coord_list[atomCounter],
                                                s1.y_coord_list[atomCounter],
                                                s1.z_coord_list[atomCounter],
                                                s1.occupancy_list[atomCounter],
                                                s1.b_factor_list[atomCounter],
                                                s1.group_list[groupIndex]['elementList'][j],
                                                s1.group_list[groupIndex]['formalChargeList'][j])

                atomCounter += 1

            # TODO not sure if we should add bonds like this
            # TODO bondAtomList == getGroupBondIndices?
            for k in range(len(s1.group_list[groupIndex]["bondOrderList"])):
                bondIndOne = s1.group_list[groupIndex]["bondAtomList"][k * 2]
                bondIndTwo = s1.group_list[groupIndex]["bondAtomList"][k * 2 + 1]
                bondOrder = s1.group_list[groupIndex]["bondOrderList"][k]
                combinedStructure.set_group_bond(
                    bondIndOne, bondIndTwo, bondOrder)

            groupCounter += 1

        # Set entity and chain info for s2
        chainToEntityIndex = self._get_chain_to_entity_index(s2)[0]
        combinedStructure.set_entity_info([1],
                                          s2.entity_list[chainToEntityIndex]['sequence'],
                                          s2.entity_list[chainToEntityIndex]['description'],
                                          s2.entity_list[chainToEntityIndex]['type'])

        combinedStructure.set_chain_info(s2.chain_id_list[0],
                                         s2.chain_name_list[0],
                                         s2.groups_per_chain[0])

        groupCounter = 0
        atomCounter = 0

        for i in range(s2.groups_per_chain[0]):
            groupIndex = s2.group_type_list[groupCounter]

            # Set group info
            combinedStructure.set_group_info(s2.group_list[groupIndex]['groupName'],
                                             s2.group_id_list[groupCounter],
                                             s2.ins_code_list[groupCounter],
                                             s2.group_list[groupIndex]['chemCompType'],
                                             len(s2.group_list[groupIndex]
                                                 ['atomNameList']),
                                             len(s2.group_list[groupIndex]
                                                 ['bondOrderList']),
                                             s2.group_list[groupIndex]['singleLetterCode'],
                                             s2.sequence_index_list[groupCounter],
                                             s2.sec_struct_list[groupCounter])

            for j in range(len(s2.group_list[groupIndex]['atomNameList'])):
                combinedStructure.set_atom_info(s2.group_list[groupIndex]['atomNameList'][j],
                                                s2.atom_id_list[atomCounter],
                                                s2.alt_loc_list[atomCounter],
                                                s2.x_coord_list[atomCounter],
                                                s2.y_coord_list[atomCounter],
                                                s2.z_coord_list[atomCounter],
                                                s2.occupancy_list[atomCounter],
                                                s2.b_factor_list[atomCounter],
                                                s2.group_list[groupIndex]['elementList'][j],
                                                s2.group_list[groupIndex]['formalChargeList'][j])

                atomCounter += 1

            # TODO not sure if we should add bonds like this
            # TODO bondAtomList == getGroupBondIndices?
            for k in range(len(s2.group_list[groupIndex]["bondOrderList"])):
                bondIndOne = s2.group_list[groupIndex]["bondAtomList"][k * 2]
                bondIndTwo = s2.group_list[groupIndex]["bondAtomList"][k * 2 + 1]
                bondOrder = s2.group_list[groupIndex]["bondOrderList"][k]
                combinedStructure.set_group_bond(
                    bondIndOne, bondIndTwo, bondOrder)

            groupCounter += 1

        combinedStructure.finalize_structure()
        return (structureId, combinedStructure)

    def _get_num_atoms_and_bonds(self, structure):
        '''Gets the number of atoms and bonds per chain
        '''
        numChains = structure.chains_per_model[0]
        atomsPerChain = [0] * numChains
        bondsPerChain = [0] * numChains
        groupCounter = 0

        for i in range(numChains):

            for j in range(structure.groups_per_chain[i]):
                groupIndex = structure.group_type_list[groupCounter]
                atomsPerChain[i] = len(
                    structure.group_list[groupIndex]['atomNameList'])
                bondsPerChain[i] = len(
                    structure.group_list[groupIndex]['bondOrderList'])
                groupCounter += 1

        return atomsPerChain, bondsPerChain

    def _get_chain_to_entity_index(self, structure):
        '''Returns an list that maps a chain index to an entity index.
        Attributes
        ----------
            structure: structureDataInterFace
        '''
        entityChainIndex = [0] * structure.num_chains

        for i in range(len(structure.entity_list)):

            for j in structure.entity_list[i]["chainIndexList"]:
                entityChainIndex[j] = i

        return entityChainIndex



#!/user/bin/env python
'''sequenceSimilarity.py

This filter returns entries that pass the sequence similarity search
criteria. Searches protein and nucleic acid sequences using the BLAST.
PSI-BLAST is used to find more distantly related protein sequences.

The E value, or Expect value, is a parameter that describes the number of
hits one can expect to see just by chance when searching a database of a
particular size. For example, an E value of one indicates that a result will
contain one sequence with similar score simply by chance. The scoring takes
chain length into consideration and therefore shorter sequences can have
identical matches with high E value.

The Low Complexity filter masks low complexity regions in a sequence to
filter out avoid spurious alignments.

Sequence Identity Cutoff (%) filter removes entries of low sequence
similarity. The cutoff value is a percentage value between 0 to 100.

Note: sequences must be at least 12 residues long. For shorter sequences try
the Sequence Motif Search.

References
----------
    BLAST:
        BLAST: Sequence searching using NCBI's BLAST (Basic Local Alignment
        Search Tool) Program , Altschul, S.F., Gish, W., Miller, W., Myers, E.W.
        and Lipman, D.J. Basic local alignment search tool. J. Mol. Biol. 215:
        403-410 (1990)
    PSI-BLAST:
        Sequence searching to detect distantly related evolutionary
        relationships using NCBI's PSI-BLAST (Position-Specific Iterated BLAST).

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from mmtfPyspark.webfilters import AdvancedQuery


class SequenceSimilarity(object):

    BLAST = 'blast'
    PSI_BLAST = 'psi-blast'

    def __init__(self, sequence, searchTool=BLAST, eValueCutoff=10.0,
                 sequenceIdentityCutoff=0, maskLowComplexity=True):
        '''Filters by squence similarity using all default parameters.

        Attribute
        ---------
            sequence (str): query sequence
            searchTool (class variable): sequenceSimilarity.BLAST or sequenceSimilarity.PSI_BLAST
            eValueCutoff (float): maximun e-value
            sequenceIdentityCutoff (int): minimum sequence identity cutoff
            maskLowComplexity (bool) : if true, mask (ignore) low complexity regions in sequence
        '''

        if len(sequence) < 12:
            raise ValueError(
                "ERROR: the query sequence must be at least 12 residues long")

        complexity = 'yes' if maskLowComplexity else 'no'

        query = "<orgPdbQuery>" + "<queryType>org.pdb.query.simple.SequenceQuery</queryType>" \
                + "<sequence>" + sequence + "</sequence>" + "<searchTool>" + searchTool \
                + "</searchTool>" + "<maskLowComplexity>" + complexity\
                + "</maskLowComplexity>" + "<eValueCutoff>" + str(eValueCutoff) \
                + "</eValueCutoff>" + "<sequenceIdentityCutoff>" + str(sequenceIdentityCutoff) \
                + "</sequenceIdentityCutoff>" + "</orgPdbQuery>"

        self.filter = AdvancedQuery(query)

    def __call__(self, t):
        return self.filter(t)



#!/usr/bin/env python
'''mmtfStructure.py

Decode msgpack unpacked data to mmtf structure

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import numpy as np
import time
import struct
from mmtf.utils import decoder_utils
from mmtfPyspark.utils import mmtfDecoder


class MmtfStructure(object):
    model_counter = 0
    chain_counter = 0
    group_counter = 0
    atom_counter = 0

    def __init__(self, input_data):
        """Decodes a msgpack unpacked data to mmtf structure
        """

        # Variables that are not in all mmtf files
        if "bFactorList" in input_data:
            int_array = np.frombuffer(input_data["bFactorList"][12:], '>i2')
            decode_num = np.frombuffer(input_data["bFactorList"][8:12], '>i')
            self.b_factor_list = mmtfDecoder.recursive_index_decode(
                int_array, decode_num)
        else:
            self.b_factor_list = []
        if 'resolution' in input_data:
            self.resolution = input_data['resolution']
        else:
            self.resolution = None
        if "rFree" in input_data:
            self.r_free = input_data["rFree"]
        else:
            self.r_free = None
        if "rWork" in input_data:
            self.r_work = input_data["rWork"]
        else:
            self.r_work = None
        if "bioAssemblyList" in input_data:
            self.bio_assembly = input_data["bioAssemblyList"]
        else:
            self.bio_assembly = []
        if "unitCell" in input_data:
            self.unit_cell = input_data["unitCell"]
        else:
            self.unit_cell = None
        if "releaseDate" in input_data:
            self.release_date = input_data["releaseDate"]
        else:
            self.release_date = None
        if "depositionDate" in input_data:
            self.deposition_date = input_data["depositionDate"]
        else:
            self.deposition_date = None
        if "title" in input_data:
            self.title = input_data["title"]
        else:
            self.title = None
        if "mmtfVersion" in input_data:
            self.mmtf_version = input_data["mmtfVersion"]
        else:
            self.mmtf_version = None
        if "mmtfProducer" in input_data:
            self.mmtf_producer = input_data["mmtfProducer"]
        else:
            self.mmtf_producer = None
        if "structureId" in input_data:
            self.structure_id = input_data["structureId"]
        else:
            self.structure_id = None
        if "spaceGroup" in input_data:
            self.space_group = input_data["spaceGroup"]
        else:
            self.space_group = None
        if "bondAtomList" in input_data:
            self.bond_atom_list = np.frombuffer(
                input_data["bondAtomList"][12:], '>i4')
        else:
            self.bond_atom_list = None
        if "bondOrderList" in input_data:
            self.bond_order_list = np.frombuffer(
                input_data["bondOrderList"][12:], '>i1')
        else:
            self.bond_order_list = None
        if "secStructList" in input_data:
            self.sec_struct_list = np.frombuffer(
                input_data["secStructList"][12:], '>i1')
        else:
            self.sec_struct_list = []
        if "atomIdList" in input_data:
            self.atom_id_list = np.cumsum(mmtfDecoder.run_length_decoder_numpy(
                np.frombuffer(input_data['atomIdList'][12:], '>i4')).astype(np.int16))
        else:
            self.atom_id_list = []
        if "sequenceIndexList" in input_data:
            self.sequence_index_list = np.cumsum(mmtfDecoder.run_length_decoder_numpy(
                np.frombuffer(input_data['sequenceIndexList'][12:], '>i4')).astype(np.int16))
        else:
            self.sequence_index_list = []
        if "occupancyList" in input_data:
            self.occupancy_list = mmtfDecoder.run_length_decoder_numpy(
                np.frombuffer(input_data["occupancyList"][12:], ">i4")) / 100
        else:
            self.occupancy_list = []
        if "experimentalMethods" in input_data:
            self.experimental_methods = input_data["experimentalMethods"]
        else:
            self.experimental_methods = None
        if "insCodeList" in input_data:
            self.ins_code_list = [chr(a) for a in mmtfDecoder.run_length_decoder_numpy(
                np.frombuffer(input_data["insCodeList"][12:], ">i4")).astype(np.int16)]
        else:
            self.ins_code_list = []
        if "entityList" in input_data:
            self.entity_list = input_data["entityList"]
        else:
            self.entity_list = []

        if "chainNameList" in input_data:
            self.chain_name_list = np.frombuffer(
                input_data["chainNameList"][12:], 'S4').astype(str)
        else:
            self.chain_name_list = []

        # Variables gaurenteed in mmtf files
        self.num_bonds = input_data["numBonds"]
        self.num_chains = input_data["numChains"]
        self.num_models = input_data["numModels"]
        self.num_atoms = input_data["numAtoms"]
        self.num_groups = input_data["numGroups"]
        self.chains_per_model = input_data["chainsPerModel"]
        self.groups_per_chain = input_data["groupsPerChain"]
        self.group_id_list = np.cumsum(mmtfDecoder.run_length_decoder_numpy(
            np.frombuffer(input_data['groupIdList'][12:], '>i4'))).astype(np.int32)
        self.group_type_list = np.frombuffer(
            input_data['groupTypeList'][12:], '>i4')
        self.x_coord_list = mmtfDecoder.recursive_index_decode(np.frombuffer(
            input_data['xCoordList'][12:], '>i2'), np.frombuffer(input_data['xCoordList'][8:12], '>i'))
        self.y_coord_list = mmtfDecoder.recursive_index_decode(np.frombuffer(
            input_data['yCoordList'][12:], '>i2'), np.frombuffer(input_data['yCoordList'][8:12], '>i'))
        self.z_coord_list = mmtfDecoder.recursive_index_decode(np.frombuffer(
            input_data['zCoordList'][12:], '>i2'), np.frombuffer(input_data['xCoordList'][8:12], '>i'))
        self.group_list = input_data['groupList']
        self.chain_id_list = np.frombuffer(
            input_data["chainIdList"][12:], 'S4').astype(str)
        self.alt_loc_list = input_data['altLocList'][12:]
        self.alt_loc_set = False

    def pass_data_on(self, data_setters):
        """Write the data from the getters to the setters.

        Attributes
        ----------
            data_setters (DataTransferInterface): a series of functions that can\
                                                  fill a chemical
        """
        self.set_alt_loc_list()
        data_setters.init_structure(self.num_bonds, len(self.x_coord_list), len(self.group_type_list),
                                    len(self.chain_id_list), len(self.chains_per_model), self.structure_id)
        decoder_utils.add_entity_info(self, data_setters)
        decoder_utils.add_atomic_information(self, data_setters)
        decoder_utils.add_header_info(self, data_setters)
        decoder_utils.add_xtalographic_info(self, data_setters)
        decoder_utils.generate_bio_assembly(self, data_setters)
        decoder_utils.add_inter_group_bonds(self, data_setters)
        data_setters.finalize_structure()

    def set_alt_loc_list(self):
        """Set the alternative location list for structure
        """
        self.alt_loc_list = [chr(x) for x in mmtfDecoder.run_length_decoder_numpy(
            np.frombuffer(self.alt_loc_list, ">i4")).astype(np.int16)]
        self.alt_loc_set = True
        return self
